#     Copyright 2022, Kay Hayen, mailto:kay.hayen@gmail.com
#
#     Python tests originally created or extracted from other peoples work. The
#     parts were too small to be protected.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
from __future__ import print_function

print("Module name is", __name__)


class SomeClass:
    pass


print("Class inside main module names its module as", repr(SomeClass.__module__))

if __name__ == "__main__":
    print("Executed as __main__:")

    import sys, os

    # The sys.argv[0] might contain ".exe", ".py" or no suffix at all.
    # Remove it, so the "diff" output is more acceptable.
    args = sys.argv[:]
    args[0] = (
        os.path.basename(args[0])
        .replace(".exe", ".py")
        .replace(".bin", ".py")
        .replace(".py", "")
    )

    print("Arguments were (stripped argv[0] suffix):", repr(args))

    # Output the flags, so we can test if we are compatible with these too.
    print("The sys.flags are:", sys.flags)



#     Copyright 2022, Kay Hayen, mailto:kay.hayen@gmail.com
#
#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
""" Standard plug-in to resolve pkg_resource actions at compile time rather than runtime.

Nuitka can detect some things that "pkg_resources" may not even be able to during
runtime, e.g. right now checking pip installed versions, is not a thing, while
some packages in their code, e.g. derive their __version__ value from that.
"""


import re

from nuitka.plugins.PluginBase import NuitkaPluginBase
from nuitka.utils.Utils import withNoDeprecationWarning


class NuitkaPluginResources(NuitkaPluginBase):
    plugin_name = "pkg-resources"
    plugin_desc = "Resolve version numbers at compile time."

    def __init__(self):
        with withNoDeprecationWarning():
            try:
                import pkg_resources
            except (ImportError, RuntimeError):
                self.pkg_resources = None
            else:
                self.pkg_resources = pkg_resources

    @staticmethod
    def isAlwaysEnabled():
        return True

    def _handleEasyInstallEntryScript(self, dist, group, name):
        module_name = None
        main_name = None

        # First try metadata, which is what the runner also does first.
        if self.metadata:
            dist = self.metadata.distribution(dist.partition("==")[0])

            for entry_point in dist.entry_points:
                if entry_point.group == group and entry_point.name == name:
                    module_name = entry_point.module
                    main_name = entry_point.attr

                    break

        if module_name is None and self.pkg_resources:
            with withNoDeprecationWarning():
                entry_point = self.pkg_resources.get_entry_info(dist, group, name)

            module_name = entry_point.module_name
            main_name = entry_point.name

        if module_name is None:
            self.sysexit(
                "Error, failed to resolve easy install entry script, is the installation broken?"
            )

        return r"""
import sys, re
sys.argv[0] = re.sub(r'(-script\.pyw?|\.exe)?$', '', sys.argv[0])
import %(module_name)s
sys.exit(%(module_name)s.%(main_name)s)
""" % {
            "module_name": module_name,
            "main_name": main_name,
        }

    def onModuleSourceCode(self, module_name, source_code):
        if module_name == "__main__":
            match = re.search(
                "\n# EASY-INSTALL-ENTRY-SCRIPT: '(.*?)','(.*?)','(.*?)'", source_code
            )

            if match is not None:
                self.info(
                    "Detected easy install entry script, compile time detecting entry point."
                )

                return self._handleEasyInstallEntryScript(*match.groups())

        return source_code



#     Copyright 2022, Kay Hayen, mailto:kay.hayen@gmail.com
#
#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
""" Comparison related codes.

Rich comparisons, "in", and "not in", also "is", and "is not", and the
"isinstance" check as used in conditions, as well as exception matching.
"""

from nuitka.containers.oset import OrderedSet
from nuitka.nodes.shapes.BuiltinTypeShapes import tshape_bool

from . import OperatorCodes
from .CodeHelpers import generateExpressionCode, pickCodeHelper
from .ErrorCodes import getErrorExitBoolCode, getReleaseCode, getReleaseCodes

specialized_cmp_helpers_set = OrderedSet(
    (
        "RICH_COMPARE_xx_OBJECT_OBJECT_OBJECT",
        "RICH_COMPARE_xx_CBOOL_OBJECT_OBJECT",
        "RICH_COMPARE_xx_NBOOL_OBJECT_OBJECT",
        "RICH_COMPARE_xx_OBJECT_OBJECT_STR",
        "RICH_COMPARE_xx_CBOOL_OBJECT_STR",
        "RICH_COMPARE_xx_NBOOL_OBJECT_STR",
        "RICH_COMPARE_xx_OBJECT_STR_OBJECT",
        "RICH_COMPARE_xx_CBOOL_STR_OBJECT",
        "RICH_COMPARE_xx_NBOOL_STR_OBJECT",
        "RICH_COMPARE_xx_OBJECT_STR_STR",
        "RICH_COMPARE_xx_CBOOL_STR_STR",
        "RICH_COMPARE_xx_NBOOL_STR_STR",
        "RICH_COMPARE_xx_OBJECT_OBJECT_UNICODE",
        "RICH_COMPARE_xx_CBOOL_OBJECT_UNICODE",
        "RICH_COMPARE_xx_NBOOL_OBJECT_UNICODE",
        "RICH_COMPARE_xx_OBJECT_UNICODE_OBJECT",
        "RICH_COMPARE_xx_CBOOL_UNICODE_OBJECT",
        "RICH_COMPARE_xx_NBOOL_UNICODE_OBJECT",
        "RICH_COMPARE_xx_OBJECT_UNICODE_UNICODE",
        "RICH_COMPARE_xx_CBOOL_UNICODE_UNICODE",
        "RICH_COMPARE_xx_NBOOL_UNICODE_UNICODE",
        "RICH_COMPARE_xx_CBOOL_OBJECT_BYTES",
        "RICH_COMPARE_xx_NBOOL_OBJECT_BYTES",
        "RICH_COMPARE_xx_OBJECT_BYTES_OBJECT",
        "RICH_COMPARE_xx_CBOOL_BYTES_OBJECT",
        "RICH_COMPARE_xx_NBOOL_BYTES_OBJECT",
        "RICH_COMPARE_xx_OBJECT_BYTES_BYTES",
        "RICH_COMPARE_xx_CBOOL_BYTES_BYTES",
        "RICH_COMPARE_xx_NBOOL_BYTES_BYTES",
        #        "RICH_COMPARE_xx_OBJECT_UNICODE",
        #        "RICH_COMPARE_xx_LONG_OBJECT",
        #        "RICH_COMPARE_xx_UNICODE_OBJECT",
        #        "RICH_COMPARE_xx_LONG_LONG",
        #        "RICH_COMPARE_xx_UNICODE_UNICODE",
        "RICH_COMPARE_xx_OBJECT_INT_INT",
        "RICH_COMPARE_xx_CBOOL_INT_INT",
        "RICH_COMPARE_xx_NBOOL_INT_INT",
        "RICH_COMPARE_xx_OBJECT_OBJECT_INT",
        "RICH_COMPARE_xx_CBOOL_OBJECT_INT",
        "RICH_COMPARE_xx_NBOOL_OBJECT_INT",
        "RICH_COMPARE_xx_OBJECT_INT_OBJECT",
        "RICH_COMPARE_xx_CBOOL_INT_OBJECT",
        "RICH_COMPARE_xx_NBOOL_INT_OBJECT",
        "RICH_COMPARE_xx_OBJECT_LONG_LONG",
        "RICH_COMPARE_xx_CBOOL_LONG_LONG",
        "RICH_COMPARE_xx_NBOOL_LONG_LONG",
        "RICH_COMPARE_xx_OBJECT_OBJECT_LONG",
        "RICH_COMPARE_xx_CBOOL_OBJECT_LONG",
        "RICH_COMPARE_xx_NBOOL_OBJECT_LONG",
        "RICH_COMPARE_xx_OBJECT_LONG_OBJECT",
        "RICH_COMPARE_xx_CBOOL_LONG_OBJECT",
        "RICH_COMPARE_xx_NBOOL_LONG_OBJECT",
        "RICH_COMPARE_xx_OBJECT_FLOAT_FLOAT",
        "RICH_COMPARE_xx_CBOOL_FLOAT_FLOAT",
        "RICH_COMPARE_xx_NBOOL_FLOAT_FLOAT",
        "RICH_COMPARE_xx_OBJECT_OBJECT_FLOAT",
        "RICH_COMPARE_xx_CBOOL_OBJECT_FLOAT",
        "RICH_COMPARE_xx_NBOOL_OBJECT_FLOAT",
        "RICH_COMPARE_xx_OBJECT_FLOAT_OBJECT",
        "RICH_COMPARE_xx_CBOOL_FLOAT_OBJECT",
        "RICH_COMPARE_xx_NBOOL_FLOAT_OBJECT",
        "RICH_COMPARE_xx_OBJECT_TUPLE_TUPLE",
        "RICH_COMPARE_xx_CBOOL_TUPLE_TUPLE",
        "RICH_COMPARE_xx_NBOOL_TUPLE_TUPLE",
        "RICH_COMPARE_xx_OBJECT_OBJECT_TUPLE",
        "RICH_COMPARE_xx_CBOOL_OBJECT_TUPLE",
        "RICH_COMPARE_xx_NBOOL_OBJECT_TUPLE",
        "RICH_COMPARE_xx_OBJECT_TUPLE_OBJECT",
        "RICH_COMPARE_xx_CBOOL_TUPLE_OBJECT",
        "RICH_COMPARE_xx_NBOOL_TUPLE_OBJECT",
        "RICH_COMPARE_xx_OBJECT_LIST_LIST",
        "RICH_COMPARE_xx_CBOOL_LIST_LIST",
        "RICH_COMPARE_xx_NBOOL_LIST_LIST",
        "RICH_COMPARE_xx_OBJECT_OBJECT_LIST",
        "RICH_COMPARE_xx_CBOOL_OBJECT_LIST",
        "RICH_COMPARE_xx_NBOOL_OBJECT_LIST",
        "RICH_COMPARE_xx_OBJECT_LIST_OBJECT",
        "RICH_COMPARE_xx_CBOOL_LIST_OBJECT",
        "RICH_COMPARE_xx_NBOOL_LIST_OBJECT",
        #        "RICH_COMPARE_xx_CBOOL_OBJECT_LONG",
        #        "RICH_COMPARE_xx_CBOOL_OBJECT_UNICODE",
        #        "RICH_COMPARE_xx_CBOOL_LONG_OBJECT",
        #        "RICH_COMPARE_xx_CBOOL_UNICODE_OBJECT",
        #        "RICH_COMPARE_xx_CBOOL_LONG_LONG",
        #        "RICH_COMPARE_xx_CBOOL_UNICODE_UNICODE",
    )
)


def getRichComparisonCode(
    to_name,
    comparator,
    left_name,
    right_name,
    left_shape,
    right_shape,
    needs_check,
    source_ref,
    emit,
    context,
):
    helper = pickCodeHelper(
        prefix="RICH_COMPARE_xx",
        suffix="",
        target_type=to_name.getCType(),
        left_shape=left_shape,
        right_shape=right_shape,
        helpers=specialized_cmp_helpers_set,
        nonspecialized=(),
        source_ref=source_ref,
    )

    # Lets patch this up here, instead of having one set per comparison operation.
    helper.helper_name = helper.helper_name.replace(
        "xx", OperatorCodes.rich_comparison_codes[comparator]
    )

    helper.emitHelperCall(
        to_name=to_name,
        arg_names=(left_name, right_name),
        ref_count=1,
        needs_check=needs_check,
        emit=emit,
        context=context,
    )


def generateComparisonExpressionCode(to_name, expression, emit, context):
    left = expression.subnode_left
    right = expression.subnode_right

    comparator = expression.getComparator()

    type_name = "PyObject *"
    if comparator in ("Is", "IsNot"):
        if left.getTypeShape() is tshape_bool and right.getTypeShape() is tshape_bool:
            type_name = "nuitka_bool"

    left_name = context.allocateTempName("cmp_expr_left", type_name=type_name)
    right_name = context.allocateTempName("cmp_expr_right", type_name=type_name)

    generateExpressionCode(
        to_name=left_name, expression=left, emit=emit, context=context
    )
    generateExpressionCode(
        to_name=right_name, expression=right, emit=emit, context=context
    )

    if comparator in OperatorCodes.containing_comparison_codes:
        needs_check = right.mayRaiseExceptionIn(BaseException, expression.subnode_left)

        res_name = context.getIntResName()

        emit(
            "%s = PySequence_Contains(%s, %s);"
            % (res_name, right_name, left_name)  # sequence goes first in the API.
        )

        getErrorExitBoolCode(
            condition="%s == -1" % res_name,
            release_names=(left_name, right_name),
            needs_check=needs_check,
            emit=emit,
            context=context,
        )

        to_name.getCType().emitAssignmentCodeFromBoolCondition(
            to_name=to_name,
            condition="%s == %d" % (res_name, 1 if comparator == "In" else 0),
            emit=emit,
        )
    elif comparator == "Is":
        to_name.getCType().emitAssignmentCodeFromBoolCondition(
            to_name=to_name, condition="%s == %s" % (left_name, right_name), emit=emit
        )

        getReleaseCodes(
            release_names=(left_name, right_name), emit=emit, context=context
        )
    elif comparator == "IsNot":
        to_name.getCType().emitAssignmentCodeFromBoolCondition(
            to_name=to_name, condition="%s != %s" % (left_name, right_name), emit=emit
        )

        getReleaseCodes(
            release_names=(left_name, right_name), emit=emit, context=context
        )
    elif comparator in OperatorCodes.rich_comparison_codes:
        getRichComparisonCode(
            to_name=to_name,
            comparator=comparator,
            left_shape=left.getTypeShape(),
            right_shape=right.getTypeShape(),
            left_name=left_name,
            right_name=right_name,
            needs_check=expression.mayRaiseExceptionComparison(),
            source_ref=expression.source_ref,
            emit=emit,
            context=context,
        )
    elif comparator in ("exception_match", "exception_mismatch"):
        needs_check = expression.mayRaiseExceptionComparison()

        res_name = context.getIntResName()

        emit("%s = EXCEPTION_MATCH_BOOL(%s, %s);" % (res_name, left_name, right_name))

        getErrorExitBoolCode(
            condition="%s == -1" % res_name,
            release_names=(left_name, right_name),
            needs_check=needs_check,
            emit=emit,
            context=context,
        )

        to_name.getCType().emitAssignmentCodeFromBoolCondition(
            to_name=to_name,
            condition="%s %s 0"
            % (res_name, "!=" if comparator == "exception_match" else "=="),
            emit=emit,
        )
    else:
        assert False, comparator


def generateBuiltinIsinstanceCode(to_name, expression, emit, context):
    inst_name = context.allocateTempName("isinstance_inst")
    cls_name = context.allocateTempName("isinstance_cls")

    generateExpressionCode(
        to_name=inst_name,
        expression=expression.subnode_instance,
        emit=emit,
        context=context,
    )
    generateExpressionCode(
        to_name=cls_name,
        expression=expression.subnode_classes,
        emit=emit,
        context=context,
    )

    context.setCurrentSourceCodeReference(expression.getCompatibleSourceReference())

    res_name = context.getIntResName()

    emit("%s = PyObject_IsInstance(%s, %s);" % (res_name, inst_name, cls_name))

    getErrorExitBoolCode(
        condition="%s == -1" % res_name,
        release_names=(inst_name, cls_name),
        emit=emit,
        context=context,
    )

    to_name.getCType().emitAssignmentCodeFromBoolCondition(
        to_name=to_name, condition="%s != 0" % res_name, emit=emit
    )


def generateBuiltinIssubclassCode(to_name, expression, emit, context):
    cls_name = context.allocateTempName("issubclass_cls")
    classes_name = context.allocateTempName("issubclass_classes")

    generateExpressionCode(
        to_name=cls_name,
        expression=expression.subnode_cls,
        emit=emit,
        context=context,
    )
    generateExpressionCode(
        to_name=classes_name,
        expression=expression.subnode_classes,
        emit=emit,
        context=context,
    )

    context.setCurrentSourceCodeReference(expression.getCompatibleSourceReference())

    res_name = context.getIntResName()

    emit("%s = PyObject_IsSubclass(%s, %s);" % (res_name, cls_name, classes_name))

    getErrorExitBoolCode(
        condition="%s == -1" % res_name,
        release_names=(cls_name, classes_name),
        emit=emit,
        context=context,
    )

    to_name.getCType().emitAssignmentCodeFromBoolCondition(
        to_name=to_name, condition="%s != 0" % res_name, emit=emit
    )


def generateTypeCheckCode(to_name, expression, emit, context):
    cls_name = context.allocateTempName("issubclass_cls")

    generateExpressionCode(
        to_name=cls_name,
        expression=expression.subnode_cls,
        emit=emit,
        context=context,
    )
    context.setCurrentSourceCodeReference(expression.getCompatibleSourceReference())

    res_name = context.getIntResName()

    emit("%s = PyType_Check(%s);" % (res_name, cls_name))

    getReleaseCode(
        release_name=cls_name,
        emit=emit,
        context=context,
    )

    to_name.getCType().emitAssignmentCodeFromBoolCondition(
        to_name=to_name, condition="%s != 0" % res_name, emit=emit
    )


def generateMatchTypeCheckMappingCode(to_name, expression, emit, context):
    cls_name = context.allocateTempName("mapping_check_cls")

    generateExpressionCode(
        to_name=cls_name,
        expression=expression.subnode_value,
        emit=emit,
        context=context,
    )

    res_name = context.getIntResName()

    emit("%s = Py_TYPE(%s)->tp_flags & Py_TPFLAGS_MAPPING;" % (res_name, cls_name))

    getReleaseCode(
        release_name=cls_name,
        emit=emit,
        context=context,
    )

    to_name.getCType().emitAssignmentCodeFromBoolCondition(
        to_name=to_name, condition="%s" % res_name, emit=emit
    )


def generateMatchTypeCheckSequenceCode(to_name, expression, emit, context):
    cls_name = context.allocateTempName("sequence_check_cls")

    generateExpressionCode(
        to_name=cls_name,
        expression=expression.subnode_value,
        emit=emit,
        context=context,
    )

    res_name = context.getIntResName()

    emit("%s = Py_TYPE(%s)->tp_flags & Py_TPFLAGS_SEQUENCE;" % (res_name, cls_name))

    getReleaseCode(
        release_name=cls_name,
        emit=emit,
        context=context,
    )

    to_name.getCType().emitAssignmentCodeFromBoolCondition(
        to_name=to_name, condition="%s" % res_name, emit=emit
    )



#     Copyright 2022, Kay Hayen, mailto:kay.hayen@gmail.com
#
#     Python test originally created or extracted from other peoples work. The
#     parts from me are licensed as below. It is at least Free Software where
#     it's copied from other people. In these cases, that will normally be
#     indicated.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
""" In this test, we get killed after some time has passed, and want to see the KeyboardInterrupt exit.

"""

# nuitka-project: --onefile

import sys
import time

print("Hello world, will you kill me?")
print(sys.version_info)
try:
    time.sleep(10)
except KeyboardInterrupt:
    print("Received KeyboardInterrupt.")
    sys.exit(0)
else:
    print("Received no KeyboardInterrupt.")
    sys.exit(1)



""" This module is only an abstraction of OrderedSet which is not present in
Python at all.

"""

# Only for re-export, pylint: disable=unused-import

try:
    from orderedset import OrderedSet
except ImportError:
    try:
        from ordered_set import OrderedSet
    except ImportError:
        from .OrderedSetFallback import OrderedSet



#     Copyright 2022, Kay Hayen, mailto:kay.hayen@gmail.com
#
#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
""" Work with ReST documentations.

This e.g. creates PDF documentations during release and tables from data
for the web site, e.g. downloads.

"""

import functools
import os
import tempfile

from .Execution import check_call, withEnvironmentVarOverridden
from .FileOperations import (
    changeFilenameExtension,
    deleteFile,
    getFileContents,
    putTextFileContents,
)


def createPDF(document):
    pdf_filename = changeFilenameExtension(document, ".pdf")
    args = ["-o", pdf_filename]

    with tempfile.NamedTemporaryFile(delete=False) as style_file:
        style_filename = style_file.name
        style_file.write(
            b"""
"pageSetup" : {
   "firstTemplate": "coverPage"
}

"styles" : [
       [ "title" , {
           "fontName": "NanumGothic-Bold",
           "fontSize": 40
       } ],
       [ "heading1" , {
           "fontName": "NanumGothic-Bold"
       } ],
       [ "heading2" , {
           "fontName": "NanumGothic"
       } ]
]
"""
        )

    if document != "Changelog.rst":
        args.append("-s")
        args.append(style_filename)

        args.append('--header="###Title### - ###Section###"')
        args.append('--footer="###Title### - page ###Page### - ###Section###"')

    # Workaround for rst2pdf not support ..code:: without language.
    old_contents = getFileContents(document)
    new_contents = old_contents.replace(".. code::\n", "::\n")

    # Add page counter reset right after TOC for PDF.
    new_contents = new_contents.replace(
        ".. contents::",
        """.. contents::

.. raw:: pdf

    PageBreak oneColumn
    SetPageCounter 1

""",
    )

    try:
        if new_contents != old_contents:
            document += ".tmp"
            putTextFileContents(filename=document, contents=new_contents)

        with withEnvironmentVarOverridden("PYTHONWARNINGS", "ignore"):
            check_call(["rst2pdf"] + args + [document])
    finally:
        if new_contents != old_contents:
            deleteFile(document, must_exist=False)

    deleteFile(style_filename, must_exist=True)

    assert os.path.exists(pdf_filename)

    return pdf_filename


def makeTable(grid):
    """Create a REST table."""

    def makeSeparator(num_cols, col_width, header_flag):
        if header_flag == 1:
            return num_cols * ("+" + (col_width) * "=") + "+\n"
        else:
            return num_cols * ("+" + (col_width) * "-") + "+\n"

    def normalizeCell(string, length):

        return string + ((length - len(string)) * " ")

    cell_width = 2 + max(
        functools.reduce(
            lambda x, y: x + y, [[len(item) for item in row] for row in grid], []
        )
    )
    num_cols = len(grid[0])
    rst = makeSeparator(num_cols, cell_width, 0)
    header_flag = 1
    for row in grid:
        rst = (
            rst
            + "| "
            + "| ".join([normalizeCell(x, cell_width - 1) for x in row])
            + "|\n"
        )
        rst = rst + makeSeparator(num_cols, cell_width, header_flag)
        header_flag = 0

    return rst



#     Copyright 2022, Kay Hayen, mailto:kay.hayen@gmail.com
#
#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
"""Str operation specs. """

from nuitka import Options

from .BuiltinParameterSpecs import (
    BuiltinParameterSpec,
    BuiltinParameterSpecNoKeywords,
)


class MethodKeywordErrorCompatibilityMixin:
    def getKeywordRefusalText(self):
        if Options.is_full_compat:
            assert "." in self.name, self.name

            try:
                eval(  # These are harmless calls, pylint: disable=eval-used
                    "''.%s(x=1)" % self.name.split(".")[-1]
                )
            except TypeError as e:
                return str(e)
            else:
                assert False, self.name
        else:
            return "%s() takes no keyword arguments" % self.name


class StrMethodSpecNoKeywords(
    MethodKeywordErrorCompatibilityMixin, BuiltinParameterSpecNoKeywords
):
    __slots__ = ()

    def __init__(
        self,
        name,
        arg_names=(),
        default_count=0,
    ):
        BuiltinParameterSpecNoKeywords.__init__(
            self,
            name="str." + name,
            arg_names=arg_names,
            default_count=default_count,
            list_star_arg=None,
            dict_star_arg=None,
            pos_only_args=(),
            kw_only_args=(),
        )


class StrMethodSpec(BuiltinParameterSpec):
    __slots__ = ()

    def __init__(
        self,
        name,
        arg_names=(),
        default_count=0,
        list_star_arg=None,
        dict_star_arg=None,
    ):
        BuiltinParameterSpec.__init__(
            self,
            name="str." + name,
            arg_names=arg_names,
            default_count=default_count,
            list_star_arg=list_star_arg,
            dict_star_arg=dict_star_arg,
            pos_only_args=(),
            kw_only_args=(),
        )


str_join_spec = StrMethodSpecNoKeywords("join", arg_names=("iterable",))
str_partition_spec = StrMethodSpecNoKeywords("partition", arg_names=("sep",))
str_rpartition_spec = StrMethodSpecNoKeywords("rpartition", arg_names=("sep",))
str_strip_spec = StrMethodSpecNoKeywords("strip", arg_names=("chars",), default_count=1)
str_lstrip_spec = StrMethodSpecNoKeywords(
    "lstrip", arg_names=("chars",), default_count=1
)
str_rstrip_spec = StrMethodSpecNoKeywords(
    "rstrip", arg_names=("chars",), default_count=1
)
str_find_spec = StrMethodSpecNoKeywords(
    "find", arg_names=("sub", "start", "end"), default_count=2
)
str_rfind_spec = StrMethodSpecNoKeywords(
    "rfind", arg_names=("sub", "start", "end"), default_count=2
)
str_index_spec = StrMethodSpecNoKeywords(
    "index", arg_names=("sub", "start", "end"), default_count=2
)
str_rindex_spec = StrMethodSpecNoKeywords(
    "rindex", arg_names=("sub", "start", "end"), default_count=2
)

str_split_spec = (StrMethodSpecNoKeywords if str is bytes else StrMethodSpec)(
    "split", arg_names=("sep", "maxsplit"), default_count=2
)
str_rsplit_spec = (StrMethodSpecNoKeywords if str is bytes else StrMethodSpec)(
    "rsplit", arg_names=("sep", "maxsplit"), default_count=2
)

str_startswith_spec = StrMethodSpecNoKeywords(
    "startswith", arg_names=("prefix", "start", "end"), default_count=2
)
str_endswith_spec = StrMethodSpecNoKeywords(
    "endswith", arg_names=("suffix", "start", "end"), default_count=2
)

str_replace_spec = StrMethodSpecNoKeywords(
    "replace", arg_names=("old", "new", "count"), default_count=1
)

str_count_spec = StrMethodSpecNoKeywords(
    "count", arg_names=("sub", "start", "end"), default_count=2
)

str_format_spec = StrMethodSpec("format", list_star_arg="args", dict_star_arg="pairs")

str_capitalize_spec = StrMethodSpecNoKeywords("capitalize", arg_names=())
str_upper_spec = StrMethodSpecNoKeywords("upper", arg_names=())
str_lower_spec = StrMethodSpecNoKeywords("lower", arg_names=())
str_swapcase_spec = StrMethodSpecNoKeywords("swapcase", arg_names=())
str_title_spec = StrMethodSpecNoKeywords("title", arg_names=())
str_isalnum_spec = StrMethodSpecNoKeywords("isalnum", arg_names=())
str_isalpha_spec = StrMethodSpecNoKeywords("isalpha", arg_names=())
str_isdigit_spec = StrMethodSpecNoKeywords("isdigit", arg_names=())
str_islower_spec = StrMethodSpecNoKeywords("islower", arg_names=())
str_isupper_spec = StrMethodSpecNoKeywords("isupper", arg_names=())
str_isspace_spec = StrMethodSpecNoKeywords("isspace", arg_names=())
str_istitle_spec = StrMethodSpecNoKeywords("istitle", arg_names=())

str_encode_spec = StrMethodSpec(
    "encode", arg_names=("encoding", "errors"), default_count=2
)

# Python2 only, Python3 this is in bytes
str_decode_spec = StrMethodSpec(
    "decode", arg_names=("encoding", "errors"), default_count=2
)

# 'center' missing.
# 'count' missing.
# 'expandtabs' missing.
# 'format' missing.
# 'ljust' missing.
# 'rjust' missing.
# 'splitlines' missing.
# 'translate' missing.
# 'zfill' missing.



#     Copyright 2022, Kay Hayen, mailto:kay.hayen@gmail.com
#
#     Python test originally created or extracted from other peoples work. The
#     parts from me are licensed as below. It is at least Free Software where
#     it's copied from other people. In these cases, that will normally be
#     indicated.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
""" This test uses the meta data lookups that wouldn't work at runtime.

But since Nuitka resolves them at compile time, no issue should happen, unless
the module used in the example is not installed.
"""

# nuitka-project: --standalone

import sys

from pkg_resources import get_distribution, require

# nuitka-skip-unless-imports: lxml

x = require("lxml")
print(x[0].version)
print(x[0].parsed_version)
__version__ = get_distribution("lxml").version
print(__version__)
__version__ = get_distribution("lxml").parsed_version
print(__version__)

try:
    from importlib.metadata import version
except ImportError:
    print("Old Python", sys.version_info)
else:
    __version__ = version("lxml")
    print(__version__)


try:
    from importlib_metadata import version
except ImportError:
    print("Backport importlib_metadata is not installed.")
else:
    __version__ = version("lxml")
    print(__version__)



#     Copyright 2022, Kay Hayen, mailto:kay.hayen@gmail.com
#
#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
""" This module is providing helper functions for complex call re-formulations.

One for each type of call. """

from nuitka.nodes.AttributeNodes import (
    ExpressionAttributeCheck,
    makeExpressionAttributeLookup,
)
from nuitka.nodes.BuiltinDictNodes import ExpressionBuiltinDict
from nuitka.nodes.BuiltinIteratorNodes import ExpressionBuiltinIter1
from nuitka.nodes.BuiltinNextNodes import ExpressionBuiltinNext1
from nuitka.nodes.BuiltinRefNodes import (
    ExpressionBuiltinAnonymousRef,
    makeExpressionBuiltinTypeRef,
)
from nuitka.nodes.BuiltinTypeNodes import ExpressionBuiltinTuple
from nuitka.nodes.CallNodes import makeExpressionCall
from nuitka.nodes.ComparisonNodes import (
    ExpressionComparisonIn,
    ExpressionComparisonIsNot,
)
from nuitka.nodes.ConditionalNodes import (
    ExpressionConditionalOr,
    makeStatementConditional,
)
from nuitka.nodes.ConstantRefNodes import makeConstantRefNode
from nuitka.nodes.ContainerMakingNodes import makeExpressionMakeTuple
from nuitka.nodes.DictionaryNodes import (
    ExpressionDictOperationIteritems,
    StatementDictOperationSet,
    StatementDictOperationSetKeyValue,
)
from nuitka.nodes.ExceptionNodes import (
    ExpressionBuiltinMakeException,
    StatementRaiseException,
)
from nuitka.nodes.FunctionAttributeNodes import ExpressionFunctionErrorStr
from nuitka.nodes.FunctionNodes import (
    ExpressionFunctionCall,
    ExpressionFunctionCreation,
    ExpressionFunctionRef,
)
from nuitka.nodes.LoopNodes import StatementLoop, StatementLoopBreak
from nuitka.nodes.OperatorNodes import makeBinaryOperationNode
from nuitka.nodes.ReturnNodes import StatementReturn
from nuitka.nodes.SubscriptNodes import (
    ExpressionSubscriptLookup,
    StatementAssignmentSubscript,
)
from nuitka.nodes.TypeNodes import (
    ExpressionBuiltinIsinstance,
    ExpressionBuiltinType1,
)
from nuitka.nodes.VariableAssignNodes import makeStatementAssignmentVariable
from nuitka.nodes.VariableDelNodes import StatementReleaseVariable
from nuitka.nodes.VariableRefNodes import (
    ExpressionTempVariableRef,
    ExpressionVariableRef,
)
from nuitka.PythonVersions import (
    getComplexCallSequenceErrorTemplate,
    python_version,
)
from nuitka.specs.ParameterSpecs import ParameterSpec

from .InternalModule import (
    internal_source_ref,
    makeInternalHelperFunctionBody,
    once_decorator,
)
from .ReformulationTryExceptStatements import makeTryExceptSingleHandlerNode
from .ReformulationTryFinallyStatements import makeTryFinallyStatement
from .TreeHelpers import (
    makeCallNode,
    makeStatementsSequenceFromStatement,
    makeStatementsSequenceFromStatements,
)

# TODO: Consider using ExpressionOutlineNodes for at least some of these
# or their own helpers.


def orderArgs(*args):
    if python_version >= 0x350:

        def weight(arg):
            result = args.index(arg)

            if arg == "kw":
                result += 1.5
            elif arg == "star_arg_list":
                result -= 1.5

            return result

        return tuple(sorted(args, key=weight))

    return args


def _makeNameAttributeLookup(node, attribute_name="__name__"):
    return makeExpressionAttributeLookup(
        expression=node, attribute_name=attribute_name, source_ref=internal_source_ref
    )


@once_decorator
def getCallableNameDescBody():
    helper_name = "get_callable_name_desc"

    # Equivalent of:
    #
    # Note: The "called_type" is a temporary variable.
    #
    # called_type = type(BuiltinFunctionType)
    #
    # if isinstance(called, (FunctionType, MethodType, BuiltinFunctionType)):
    #     return called.__name__
    # elif python_version < 0x3 and isinstance(called, ClassType):
    #     return called_type.__name__ + " constructor"
    # elif python_version < 0x3 and isinstance(called, InstanceType):
    #     return called_type.__name__ + " instance"
    # else:
    #     return called_type.__name__ + " object"

    result = makeInternalHelperFunctionBody(
        name=helper_name,
        parameters=ParameterSpec(
            ps_name=helper_name,
            ps_normal_args=("called",),
            ps_list_star_arg=None,
            ps_dict_star_arg=None,
            ps_default_count=0,
            ps_kw_only_args=(),
            ps_pos_only_args=(),
        ),
    )

    called_variable = result.getVariableForAssignment(variable_name="called")

    if python_version < 0x390:
        function_name = makeBinaryOperationNode(
            operator="Add",
            left=_makeNameAttributeLookup(
                node=ExpressionVariableRef(
                    variable=called_variable, source_ref=internal_source_ref
                ),
                attribute_name="__name__",
            ),
            right=makeConstantRefNode(
                constant="()",
                source_ref=internal_source_ref,
                user_provided=True,
            ),
            source_ref=internal_source_ref,
        )
    else:
        # TODO: Make it usable for pre-Python 3.9 too.
        function_name = ExpressionFunctionErrorStr(
            value=ExpressionVariableRef(
                variable=called_variable, source_ref=internal_source_ref
            ),
            source_ref=internal_source_ref,
        )

    functions_case = makeStatementsSequenceFromStatement(
        statement=(
            StatementReturn(
                expression=function_name,
                source_ref=internal_source_ref,
            )
        )
    )

    if python_version >= 0x390:
        result.setChild("body", functions_case)

        return result

    no_branch = StatementReturn(
        expression=makeBinaryOperationNode(
            operator="Add",
            right=makeConstantRefNode(
                constant=" object", source_ref=internal_source_ref, user_provided=True
            ),
            left=_makeNameAttributeLookup(
                ExpressionBuiltinType1(
                    value=ExpressionVariableRef(
                        variable=called_variable, source_ref=internal_source_ref
                    ),
                    source_ref=internal_source_ref,
                )
            ),
            source_ref=internal_source_ref,
        ),
        source_ref=internal_source_ref,
    )

    if python_version < 0x300:
        instance_case = StatementReturn(
            expression=makeBinaryOperationNode(
                operator="Add",
                right=makeConstantRefNode(
                    constant=" instance",
                    source_ref=internal_source_ref,
                    user_provided=True,
                ),
                left=_makeNameAttributeLookup(
                    _makeNameAttributeLookup(
                        ExpressionVariableRef(
                            variable=called_variable, source_ref=internal_source_ref
                        ),
                        attribute_name="__class__",
                    )
                ),
                source_ref=internal_source_ref,
            ),
            source_ref=internal_source_ref,
        )

        no_branch = makeStatementConditional(
            condition=ExpressionBuiltinIsinstance(
                instance=ExpressionVariableRef(
                    variable=called_variable, source_ref=internal_source_ref
                ),
                classes=ExpressionBuiltinAnonymousRef(
                    builtin_name="instance", source_ref=internal_source_ref
                ),
                source_ref=internal_source_ref,
            ),
            yes_branch=instance_case,
            no_branch=no_branch,
            source_ref=internal_source_ref,
        )

        class_case = StatementReturn(
            expression=makeBinaryOperationNode(
                operator="Add",
                right=makeConstantRefNode(
                    constant=" constructor",
                    source_ref=internal_source_ref,
                    user_provided=True,
                ),
                left=_makeNameAttributeLookup(
                    ExpressionVariableRef(
                        variable=called_variable, source_ref=internal_source_ref
                    )
                ),
                source_ref=internal_source_ref,
            ),
            source_ref=internal_source_ref,
        )

        no_branch = makeStatementConditional(
            condition=ExpressionBuiltinIsinstance(
                instance=ExpressionVariableRef(
                    variable=called_variable, source_ref=internal_source_ref
                ),
                classes=ExpressionBuiltinAnonymousRef(
                    builtin_name="classobj", source_ref=internal_source_ref
                ),
                source_ref=internal_source_ref,
            ),
            yes_branch=class_case,
            no_branch=no_branch,
            source_ref=internal_source_ref,
        )

    if python_version < 0x300:
        normal_cases = ("function", "builtin_function_or_method", "instancemethod")
    else:
        normal_cases = ("function", "builtin_function_or_method")

    result.setChild(
        "body",
        makeStatementsSequenceFromStatement(
            statement=makeStatementConditional(
                condition=ExpressionBuiltinIsinstance(
                    instance=ExpressionVariableRef(
                        variable=called_variable, source_ref=internal_source_ref
                    ),
                    classes=makeExpressionMakeTuple(
                        elements=tuple(
                            ExpressionBuiltinAnonymousRef(
                                builtin_name=builtin_name,
                                source_ref=internal_source_ref,
                            )
                            for builtin_name in normal_cases
                        ),
                        source_ref=internal_source_ref,
                    ),
                    source_ref=internal_source_ref,
                ),
                yes_branch=functions_case,
                no_branch=no_branch,
                source_ref=internal_source_ref,
            )
        ),
    )

    return result


def makeStarListArgumentErrorRaise(called_variable, star_list_variable):
    return StatementRaiseException(
        exception_type=ExpressionBuiltinMakeException(
            exception_name="TypeError",
            args=(
                makeBinaryOperationNode(
                    operator="Mod",
                    left=makeConstantRefNode(
                        constant=getComplexCallSequenceErrorTemplate(),
                        source_ref=internal_source_ref,
                        user_provided=True,
                    ),
                    right=makeExpressionMakeTuple(
                        elements=(
                            ExpressionFunctionCall(
                                function=ExpressionFunctionCreation(
                                    function_ref=ExpressionFunctionRef(
                                        function_body=getCallableNameDescBody(),
                                        source_ref=internal_source_ref,
                                    ),
                                    defaults=(),
                                    kw_defaults=None,
                                    annotations=None,
                                    source_ref=internal_source_ref,
                                ),
                                values=(
                                    ExpressionVariableRef(
                                        variable=called_variable,
                                        source_ref=internal_source_ref,
                                    ),
                                ),
                                source_ref=internal_source_ref,
                            ),
                            _makeNameAttributeLookup(
                                ExpressionBuiltinType1(
                                    value=ExpressionVariableRef(
                                        variable=star_list_variable,
                                        source_ref=internal_source_ref,
                                    ),
                                    source_ref=internal_source_ref,
                                )
                            ),
                        ),
                        source_ref=internal_source_ref,
                    ),
                    source_ref=internal_source_ref,
                ),
            ),
            source_ref=internal_source_ref,
        ),
        exception_value=None,
        exception_trace=None,
        exception_cause=None,
        source_ref=internal_source_ref,
    )


def _makeStarListArgumentToTupleStatement(called_variable, star_list_variable):
    if python_version >= 0x350:
        non_tuple_code = makeStatementConditional(
            condition=ExpressionConditionalOr(
                left=ExpressionAttributeCheck(
                    expression=ExpressionVariableRef(
                        variable=star_list_variable, source_ref=internal_source_ref
                    ),
                    attribute_name="__iter__",
                    source_ref=internal_source_ref,
                ),
                right=ExpressionAttributeCheck(
                    expression=ExpressionVariableRef(
                        variable=star_list_variable, source_ref=internal_source_ref
                    ),
                    attribute_name="__getitem__",
                    source_ref=internal_source_ref,
                ),
                source_ref=internal_source_ref,
            ),
            yes_branch=makeStatementAssignmentVariable(
                variable=star_list_variable,
                source=ExpressionBuiltinTuple(
                    value=ExpressionVariableRef(
                        variable=star_list_variable, source_ref=internal_source_ref
                    ),
                    source_ref=internal_source_ref,
                ),
                source_ref=internal_source_ref,
            ),
            no_branch=makeStarListArgumentErrorRaise(
                called_variable=called_variable, star_list_variable=star_list_variable
            ),
            source_ref=internal_source_ref,
        )
    else:
        non_tuple_code = makeTryExceptSingleHandlerNode(
            tried=makeStatementAssignmentVariable(
                variable=star_list_variable,
                source=ExpressionBuiltinTuple(
                    value=ExpressionVariableRef(
                        variable=star_list_variable, source_ref=internal_source_ref
                    ),
                    source_ref=internal_source_ref,
                ),
                source_ref=internal_source_ref,
            ),
            exception_name="TypeError",
            handler_body=makeStarListArgumentErrorRaise(
                called_variable=called_variable, star_list_variable=star_list_variable
            ),
            source_ref=internal_source_ref,
        )

    return makeStatementConditional(
        condition=ExpressionComparisonIsNot(
            left=ExpressionBuiltinType1(
                value=ExpressionVariableRef(
                    variable=star_list_variable, source_ref=internal_source_ref
                ),
                source_ref=internal_source_ref,
            ),
            right=makeExpressionBuiltinTypeRef(
                builtin_name="tuple", source_ref=internal_source_ref
            ),
            source_ref=internal_source_ref,
        ),
        yes_branch=non_tuple_code,
        no_branch=None,
        source_ref=internal_source_ref,
    )


def _makeRaiseExceptionMustBeMapping(called_variable, star_dict_variable):
    return StatementRaiseException(
        exception_type=ExpressionBuiltinMakeException(
            exception_name="TypeError",
            args=(
                makeBinaryOperationNode(
                    operator="Mod",
                    left=makeConstantRefNode(
                        constant="""\
%s argument after ** must be a mapping, not %s""",
                        source_ref=internal_source_ref,
                        user_provided=True,
                    ),
                    right=makeExpressionMakeTuple(
                        elements=(
                            ExpressionFunctionCall(
                                function=ExpressionFunctionCreation(
                                    function_ref=ExpressionFunctionRef(
                                        function_body=getCallableNameDescBody(),
                                        source_ref=internal_source_ref,
                                    ),
                                    defaults=(),
                                    kw_defaults=None,
                                    annotations=None,
                                    source_ref=internal_source_ref,
                                ),
                                values=(
                                    ExpressionVariableRef(
                                        variable=called_variable,
                                        source_ref=internal_source_ref,
                                    ),
                                ),
                                source_ref=internal_source_ref,
                            ),
                            _makeNameAttributeLookup(
                                ExpressionBuiltinType1(
                                    value=ExpressionVariableRef(
                                        variable=star_dict_variable,
                                        source_ref=internal_source_ref,
                                    ),
                                    source_ref=internal_source_ref,
                                )
                            ),
                        ),
                        source_ref=internal_source_ref,
                    ),
                    source_ref=internal_source_ref,
                ),
            ),
            source_ref=internal_source_ref,
        ),
        exception_value=None,
        exception_trace=None,
        exception_cause=None,
        source_ref=internal_source_ref,
    )


def _makeIteratingLoopStatement(tmp_iter_variable, tmp_item_variable, statements):
    loop_body = makeStatementsSequenceFromStatements(
        makeTryExceptSingleHandlerNode(
            tried=makeStatementAssignmentVariable(
                variable=tmp_item_variable,
                source=ExpressionBuiltinNext1(
                    value=ExpressionTempVariableRef(
                        variable=tmp_iter_variable, source_ref=internal_source_ref
                    ),
                    source_ref=internal_source_ref,
                ),
                source_ref=internal_source_ref,
            ),
            exception_name="StopIteration",
            handler_body=StatementLoopBreak(source_ref=internal_source_ref),
            source_ref=internal_source_ref,
        ),
        *statements
    )

    return StatementLoop(loop_body=loop_body, source_ref=internal_source_ref)


def _makeStarDictArgumentToDictStatement(result, called_variable, star_dict_variable):
    temp_scope = result.allocateTempScope("mapping")

    tmp_dict_variable = result.allocateTempVariable(temp_scope, "dict")
    tmp_iter_variable = result.allocateTempVariable(temp_scope, "iter")
    tmp_keys_variable = result.allocateTempVariable(temp_scope, "keys")
    tmp_key_variable = result.allocateTempVariable(temp_scope, "key")

    loop_body = []

    if python_version >= 0x380:
        loop_body.append(
            makeStatementConditional(
                condition=ExpressionComparisonIn(
                    left=ExpressionTempVariableRef(
                        variable=tmp_key_variable, source_ref=internal_source_ref
                    ),
                    right=ExpressionVariableRef(
                        variable=tmp_dict_variable, source_ref=internal_source_ref
                    ),
                    source_ref=internal_source_ref,
                ),
                yes_branch=_makeRaiseDuplicationItem(
                    called_variable=called_variable, tmp_key_variable=tmp_key_variable
                ),
                no_branch=None,
                source_ref=internal_source_ref,
            )
        )

    loop_body.append(
        StatementDictOperationSet(
            dict_arg=ExpressionTempVariableRef(
                variable=tmp_dict_variable, source_ref=internal_source_ref
            ),
            key=ExpressionTempVariableRef(
                variable=tmp_key_variable, source_ref=internal_source_ref
            ),
            value=ExpressionSubscriptLookup(
                expression=ExpressionVariableRef(
                    variable=star_dict_variable, source_ref=internal_source_ref
                ),
                subscript=ExpressionTempVariableRef(
                    variable=tmp_key_variable, source_ref=internal_source_ref
                ),
                source_ref=internal_source_ref,
            ),
            source_ref=internal_source_ref,
        )
    )

    mapping_case = makeStatementsSequenceFromStatements(
        makeTryExceptSingleHandlerNode(
            tried=makeStatementAssignmentVariable(
                variable=tmp_keys_variable,
                source=makeCallNode(
                    _makeNameAttributeLookup(
                        ExpressionVariableRef(
                            variable=star_dict_variable, source_ref=internal_source_ref
                        ),
                        attribute_name="keys",
                    ),
                    internal_source_ref,
                ),
                source_ref=internal_source_ref,
            ),
            exception_name="AttributeError",
            handler_body=_makeRaiseExceptionMustBeMapping(
                called_variable=called_variable, star_dict_variable=star_dict_variable
            ),
            source_ref=internal_source_ref,
        ),
        makeStatementAssignmentVariable(
            variable=tmp_iter_variable,
            source=ExpressionBuiltinIter1(
                value=ExpressionTempVariableRef(
                    variable=tmp_keys_variable, source_ref=internal_source_ref
                ),
                source_ref=internal_source_ref,
            ),
            source_ref=internal_source_ref,
        ),
        makeStatementAssignmentVariable(
            variable=tmp_dict_variable,
            source=makeConstantRefNode(
                constant={}, source_ref=internal_source_ref, user_provided=True
            ),
            source_ref=internal_source_ref,
        ),
        _makeIteratingLoopStatement(
            tmp_iter_variable=tmp_iter_variable,
            tmp_item_variable=tmp_key_variable,
            statements=loop_body,
        ),
        makeStatementAssignmentVariable(
            variable=star_dict_variable,
            source=ExpressionTempVariableRef(
                variable=tmp_dict_variable, source_ref=internal_source_ref
            ),
            source_ref=internal_source_ref,
        ),
    )

    tried = makeStatementConditional(
        condition=ExpressionComparisonIsNot(
            left=ExpressionBuiltinType1(
                value=ExpressionVariableRef(
                    variable=star_dict_variable, source_ref=internal_source_ref
                ),
                source_ref=internal_source_ref,
            ),
            right=makeExpressionBuiltinTypeRef(
                builtin_name="dict", source_ref=internal_source_ref
            ),
            source_ref=internal_source_ref,
        ),
        yes_branch=mapping_case,
        no_branch=None,
        source_ref=internal_source_ref,
    )

    final = (
        StatementReleaseVariable(
            variable=tmp_dict_variable, source_ref=internal_source_ref
        ),
        StatementReleaseVariable(
            variable=tmp_iter_variable, source_ref=internal_source_ref
        ),
        StatementReleaseVariable(
            variable=tmp_keys_variable, source_ref=internal_source_ref
        ),
        StatementReleaseVariable(
            variable=tmp_key_variable, source_ref=internal_source_ref
        ),
    )

    return makeTryFinallyStatement(
        provider=result, tried=tried, final=final, source_ref=internal_source_ref
    )


def _makeRaiseNoStringItem(called_variable):
    if python_version < 0x390:
        raise_arg = makeBinaryOperationNode(
            operator="Mod",
            left=makeConstantRefNode(
                constant="%s keywords must be strings",
                source_ref=internal_source_ref,
                user_provided=True,
            ),
            right=ExpressionFunctionCall(
                function=ExpressionFunctionCreation(
                    function_ref=ExpressionFunctionRef(
                        function_body=getCallableNameDescBody(),
                        source_ref=internal_source_ref,
                    ),
                    defaults=(),
                    kw_defaults=None,
                    annotations=None,
                    source_ref=internal_source_ref,
                ),
                values=(
                    ExpressionVariableRef(
                        variable=called_variable, source_ref=internal_source_ref
                    ),
                ),
                source_ref=internal_source_ref,
            ),
            source_ref=internal_source_ref,
        )
    else:
        raise_arg = makeConstantRefNode(
            constant="keywords must be strings",
            source_ref=internal_source_ref,
            user_provided=True,
        )

    return StatementRaiseException(
        exception_type=ExpressionBuiltinMakeException(
            exception_name="TypeError",
            args=(raise_arg,),
            source_ref=internal_source_ref,
        ),
        exception_value=None,
        exception_trace=None,
        exception_cause=None,
        source_ref=internal_source_ref,
    )


def _makeRaiseDuplicationItem(called_variable, tmp_key_variable):
    return StatementRaiseException(
        exception_type=ExpressionBuiltinMakeException(
            exception_name="TypeError",
            args=(
                makeBinaryOperationNode(
                    operator="Mod",
                    left=makeConstantRefNode(
                        constant="""\
%s got multiple values for keyword argument '%s'""",
                        source_ref=internal_source_ref,
                        user_provided=True,
                    ),
                    right=makeExpressionMakeTuple(
                        elements=(
                            ExpressionFunctionCall(
                                function=ExpressionFunctionCreation(
                                    function_ref=ExpressionFunctionRef(
                                        function_body=getCallableNameDescBody(),
                                        source_ref=internal_source_ref,
                                    ),
                                    defaults=(),
                                    kw_defaults=None,
                                    annotations=None,
                                    source_ref=internal_source_ref,
                                ),
                                values=(
                                    ExpressionVariableRef(
                                        variable=called_variable,
                                        source_ref=internal_source_ref,
                                    ),
                                ),
                                source_ref=internal_source_ref,
                            ),
                            ExpressionTempVariableRef(
                                variable=tmp_key_variable,
                                source_ref=internal_source_ref,
                            ),
                        ),
                        source_ref=internal_source_ref,
                    ),
                    source_ref=internal_source_ref,
                ),
            ),
            source_ref=internal_source_ref,
        ),
        exception_value=None,
        exception_trace=None,
        exception_cause=None,
        source_ref=internal_source_ref,
    )


def _makeStarDictArgumentMergeToKwStatement(
    result, called_variable, kw_variable, star_dict_variable
):
    # This is plain terribly complex
    temp_scope = result.allocateTempScope("dict")

    tmp_iter_variable = result.allocateTempVariable(temp_scope, "iter")
    tmp_keys_variable = result.allocateTempVariable(temp_scope, "keys")
    tmp_key_variable = result.allocateTempVariable(temp_scope, "key_xxx")

    final = [
        StatementReleaseVariable(
            variable=tmp_iter_variable, source_ref=internal_source_ref
        ),
        StatementReleaseVariable(
            variable=tmp_keys_variable, source_ref=internal_source_ref
        ),
        StatementReleaseVariable(
            variable=tmp_key_variable, source_ref=internal_source_ref
        ),
    ]

    mapping_loop_body = (
        makeStatementConditional(
            condition=ExpressionComparisonIn(
                left=ExpressionTempVariableRef(
                    variable=tmp_key_variable, source_ref=internal_source_ref
                ),
                right=ExpressionVariableRef(
                    variable=kw_variable, source_ref=internal_source_ref
                ),
                source_ref=internal_source_ref,
            ),
            yes_branch=_makeRaiseDuplicationItem(
                called_variable=called_variable, tmp_key_variable=tmp_key_variable
            ),
            no_branch=None,
            source_ref=internal_source_ref,
        ),
        StatementAssignmentSubscript(
            subscribed=ExpressionVariableRef(
                variable=kw_variable, source_ref=internal_source_ref
            ),
            subscript=ExpressionTempVariableRef(
                variable=tmp_key_variable, source_ref=internal_source_ref
            ),
            source=ExpressionSubscriptLookup(
                expression=ExpressionVariableRef(
                    variable=star_dict_variable, source_ref=internal_source_ref
                ),
                subscript=ExpressionTempVariableRef(
                    variable=tmp_key_variable, source_ref=internal_source_ref
                ),
                source_ref=internal_source_ref,
            ),
            source_ref=internal_source_ref,
        ),
    )

    mapping_case = makeStatementsSequenceFromStatements(
        makeTryExceptSingleHandlerNode(
            tried=makeStatementAssignmentVariable(
                variable=tmp_keys_variable,
                source=makeCallNode(
                    _makeNameAttributeLookup(
                        ExpressionVariableRef(
                            variable=star_dict_variable, source_ref=internal_source_ref
                        ),
                        attribute_name="keys",
                    ),
                    internal_source_ref,
                ),
                source_ref=internal_source_ref,
            ),
            exception_name="AttributeError",
            handler_body=_makeRaiseExceptionMustBeMapping(
                called_variable=called_variable, star_dict_variable=star_dict_variable
            ),
            source_ref=internal_source_ref,
        ),
        makeStatementAssignmentVariable(
            variable=tmp_iter_variable,
            source=ExpressionBuiltinIter1(
                value=ExpressionTempVariableRef(
                    variable=tmp_keys_variable, source_ref=internal_source_ref
                ),
                source_ref=internal_source_ref,
            ),
            source_ref=internal_source_ref,
        ),
        _makeIteratingLoopStatement(
            tmp_iter_variable=tmp_iter_variable,
            tmp_item_variable=tmp_key_variable,
            statements=mapping_loop_body,
        ),
    )

    temp_scope = result.allocateTempScope("dict")

    tmp_iter_variable = result.allocateTempVariable(temp_scope, "iter")
    tmp_item_variable = result.allocateTempVariable(temp_scope, "item")
    tmp_key_variable = result.allocateTempVariable(temp_scope, "key")

    final += (
        StatementReleaseVariable(
            variable=tmp_iter_variable, source_ref=internal_source_ref
        ),
        StatementReleaseVariable(
            variable=tmp_item_variable, source_ref=internal_source_ref
        ),
        StatementReleaseVariable(
            variable=tmp_key_variable, source_ref=internal_source_ref
        ),
    )

    dict_loop_body = (
        makeStatementAssignmentVariable(
            variable=tmp_key_variable,
            source=ExpressionSubscriptLookup(
                expression=ExpressionTempVariableRef(
                    variable=tmp_item_variable, source_ref=internal_source_ref
                ),
                subscript=makeConstantRefNode(
                    constant=0, source_ref=internal_source_ref, user_provided=True
                ),
                source_ref=internal_source_ref,
            ),
            source_ref=internal_source_ref,
        ),
        makeStatementConditional(
            condition=ExpressionComparisonIn(
                left=ExpressionTempVariableRef(
                    variable=tmp_key_variable, source_ref=internal_source_ref
                ),
                right=ExpressionVariableRef(
                    variable=kw_variable, source_ref=internal_source_ref
                ),
                source_ref=internal_source_ref,
            ),
            yes_branch=_makeRaiseDuplicationItem(
                called_variable=called_variable, tmp_key_variable=tmp_key_variable
            ),
            no_branch=None,
            source_ref=internal_source_ref,
        ),
        StatementAssignmentSubscript(
            subscribed=ExpressionVariableRef(
                variable=kw_variable, source_ref=internal_source_ref
            ),
            subscript=ExpressionTempVariableRef(
                variable=tmp_key_variable, source_ref=internal_source_ref
            ),
            source=ExpressionSubscriptLookup(
                expression=ExpressionTempVariableRef(
                    variable=tmp_item_variable, source_ref=internal_source_ref
                ),
                subscript=makeConstantRefNode(
                    constant=1, source_ref=internal_source_ref, user_provided=True
                ),
                source_ref=internal_source_ref,
            ),
            source_ref=internal_source_ref,
        ),
    )

    dict_case = makeStatementsSequenceFromStatements(
        makeStatementAssignmentVariable(
            variable=kw_variable,
            source=ExpressionBuiltinDict(
                pos_arg=ExpressionVariableRef(
                    variable=kw_variable, source_ref=internal_source_ref
                ),
                pairs=(),
                source_ref=internal_source_ref,
            ),
            source_ref=internal_source_ref,
        ),
        makeStatementAssignmentVariable(
            variable=tmp_iter_variable,
            source=ExpressionBuiltinIter1(
                value=ExpressionDictOperationIteritems(
                    dict_arg=ExpressionVariableRef(
                        variable=star_dict_variable, source_ref=internal_source_ref
                    ),
                    source_ref=internal_source_ref,
                ),
                source_ref=internal_source_ref,
            ),
            source_ref=internal_source_ref,
        ),
        _makeIteratingLoopStatement(
            tmp_iter_variable=tmp_iter_variable,
            tmp_item_variable=tmp_item_variable,
            statements=dict_loop_body,
        ),
    )

    dict_case = makeStatementConditional(
        condition=ExpressionVariableRef(
            variable=star_dict_variable, source_ref=internal_source_ref
        ),
        yes_branch=dict_case,
        no_branch=None,
        source_ref=internal_source_ref,
    )

    tried = makeStatementConditional(
        condition=ExpressionComparisonIsNot(
            left=ExpressionBuiltinType1(
                value=ExpressionVariableRef(
                    variable=star_dict_variable, source_ref=internal_source_ref
                ),
                source_ref=internal_source_ref,
            ),
            right=makeExpressionBuiltinTypeRef(
                builtin_name="dict", source_ref=internal_source_ref
            ),
            source_ref=internal_source_ref,
        ),
        yes_branch=mapping_case,
        no_branch=dict_case,
        source_ref=internal_source_ref,
    )

    return makeTryFinallyStatement(
        provider=result, tried=tried, final=final, source_ref=internal_source_ref
    )


@once_decorator
def getFunctionCallHelperStarList():
    helper_name = "complex_call_helper_star_list"
    # Equivalent of:
    #
    # Note: Call in here is not the same, as it can go without checks directly
    # to PyObject_Call.
    #
    # if not isinstance(star_arg_list, tuple):
    #     try:
    #         star_arg_list = tuple(star_arg_list)
    #     except TypeError:
    #         raise TypeError, "%s argument after * must be a sequence, not %s" % (
    #             get_callable_name_desc(function),
    #             type(star_arg_list).__name__
    #         )
    #
    # return called(*star_arg_list)

    # Only need to check if the star argument value is a sequence and then
    # convert to tuple.
    result = makeInternalHelperFunctionBody(
        name=helper_name,
        parameters=ParameterSpec(
            ps_name=helper_name,
            ps_normal_args=("called", "star_arg_list"),
            ps_list_star_arg=None,
            ps_dict_star_arg=None,
            ps_default_count=0,
            ps_kw_only_args=(),
            ps_pos_only_args=(),
        ),
    )

    called_variable = result.getVariableForAssignment(variable_name="called")
    star_arg_list_variable = result.getVariableForAssignment(
        variable_name="star_arg_list"
    )

    body = makeStatementsSequenceFromStatements(
        _makeStarListArgumentToTupleStatement(
            called_variable=called_variable, star_list_variable=star_arg_list_variable
        ),
        StatementReturn(
            expression=makeExpressionCall(
                called=ExpressionVariableRef(
                    variable=called_variable, source_ref=internal_source_ref
                ),
                args=ExpressionVariableRef(
                    variable=star_arg_list_variable, source_ref=internal_source_ref
                ),
                kw=None,
                source_ref=internal_source_ref,
            ),
            source_ref=internal_source_ref,
        ),
    )

    result.setChild("body", body)

    return result


@once_decorator
def getFunctionCallHelperKeywordsStarList():
    helper_name = "complex_call_helper_keywords_star_list"

    # Equivalent of:
    #
    # Note: Call in here is not the same, as it can go without checks directly
    # to PyObject_Call.
    #
    # if not isinstance(star_arg_list, tuple):
    #     try:
    #         star_arg_list = tuple(star_arg_list)
    #     except TypeError:
    #         raise TypeError, "%s argument after * must be a sequence, not %s" % (
    #             get_callable_name_desc(function),
    #             type(star_arg_list).__name__
    #         )
    #
    # return called(*star_arg_list)

    # Only need to check if the star argument value is a sequence and then
    # convert to tuple.
    result = makeInternalHelperFunctionBody(
        name=helper_name,
        parameters=ParameterSpec(
            ps_name=helper_name,
            ps_normal_args=orderArgs("called", "kw", "star_arg_list"),
            ps_list_star_arg=None,
            ps_dict_star_arg=None,
            ps_default_count=0,
            ps_kw_only_args=(),
            ps_pos_only_args=(),
        ),
    )

    called_variable = result.getVariableForAssignment(variable_name="called")

    kw_variable = result.getVariableForAssignment(variable_name="kw")

    star_arg_list_variable = result.getVariableForAssignment(
        variable_name="star_arg_list"
    )

    body = makeStatementsSequenceFromStatements(
        _makeStarListArgumentToTupleStatement(
            called_variable=called_variable, star_list_variable=star_arg_list_variable
        ),
        StatementReturn(
            expression=makeExpressionCall(
                called=ExpressionVariableRef(
                    variable=called_variable, source_ref=internal_source_ref
                ),
                args=ExpressionVariableRef(
                    variable=star_arg_list_variable, source_ref=internal_source_ref
                ),
                kw=ExpressionVariableRef(
                    variable=kw_variable, source_ref=internal_source_ref
                ),
                source_ref=internal_source_ref,
            ),
            source_ref=internal_source_ref,
        ),
    )

    result.setChild("body", body)

    return result


@once_decorator
def getFunctionCallHelperPosStarList():
    helper_name = "complex_call_helper_pos_star_list"

    # Equivalent of:
    #
    # Note: Call in here is not the same, as it can go without checks directly
    # to PyObject_Call.
    #
    # if not isinstance(star_arg_list, tuple):
    #     try:
    #         star_arg_list = tuple(star_arg_list)
    #     except TypeError:
    #         raise TypeError, "%s argument after * must be a sequence, not %s" % (
    #             get_callable_name_desc(function),
    #             type(star_arg_list).__name__
    #         )
    #
    # return called(*star_arg_list)

    # Only need to check if the star argument value is a sequence and then
    # convert to tuple.
    result = makeInternalHelperFunctionBody(
        name=helper_name,
        parameters=ParameterSpec(
            ps_name=helper_name,
            ps_normal_args=("called", "args", "star_arg_list"),
            ps_list_star_arg=None,
            ps_dict_star_arg=None,
            ps_default_count=0,
            ps_kw_only_args=(),
            ps_pos_only_args=(),
        ),
    )

    called_variable = result.getVariableForAssignment(variable_name="called")

    args_variable = result.getVariableForAssignment(variable_name="args")

    star_arg_list_variable = result.getVariableForAssignment(
        variable_name="star_arg_list"
    )

    body = makeStatementsSequenceFromStatements(
        _makeStarListArgumentToTupleStatement(
            called_variable=called_variable, star_list_variable=star_arg_list_variable
        ),
        StatementReturn(
            expression=makeExpressionCall(
                called=ExpressionVariableRef(
                    variable=called_variable, source_ref=internal_source_ref
                ),
                args=makeBinaryOperationNode(
                    operator="Add",
                    left=ExpressionVariableRef(
                        variable=args_variable, source_ref=internal_source_ref
                    ),
                    right=ExpressionVariableRef(
                        variable=star_arg_list_variable, source_ref=internal_source_ref
                    ),
                    source_ref=internal_source_ref,
                ),
                kw=None,
                source_ref=internal_source_ref,
            ),
            source_ref=internal_source_ref,
        ),
    )

    result.setChild("body", body)

    return result


@once_decorator
def getFunctionCallHelperPosKeywordsStarList():
    helper_name = "complex_call_helper_pos_keywords_star_list"
    # Equivalent of:
    #
    # Note: Call in here is not the same, as it can go without checks directly
    # to PyObject_Call.
    #
    # if not isinstance(star_arg_list, tuple):
    #     try:
    #         star_arg_list = tuple(star_arg_list)
    #     except TypeError:
    #         raise TypeError, "%s argument after * must be a sequence, not %s" % (
    #             get_callable_name_desc(function),
    #             type(star_arg_list).__name__
    #         )
    #
    # return called(*star_arg_list)

    # Only need to check if the star argument value is a sequence and then
    # convert to tuple.
    result = makeInternalHelperFunctionBody(
        name=helper_name,
        parameters=ParameterSpec(
            ps_name=helper_name,
            ps_normal_args=orderArgs("called", "args", "kw", "star_arg_list"),
            ps_list_star_arg=None,
            ps_dict_star_arg=None,
            ps_default_count=0,
            ps_kw_only_args=(),
            ps_pos_only_args=(),
        ),
    )

    called_variable = result.getVariableForAssignment(variable_name="called")

    args_variable = result.getVariableForAssignment(variable_name="args")

    kw_variable = result.getVariableForAssignment(variable_name="kw")

    star_arg_list_variable = result.getVariableForAssignment(
        variable_name="star_arg_list"
    )

    body = makeStatementsSequenceFromStatements(
        _makeStarListArgumentToTupleStatement(
            called_variable=called_variable, star_list_variable=star_arg_list_variable
        ),
        StatementReturn(
            expression=makeExpressionCall(
                called=ExpressionVariableRef(
                    variable=called_variable, source_ref=internal_source_ref
                ),
                args=makeBinaryOperationNode(
                    operator="Add",
                    left=ExpressionVariableRef(
                        variable=args_variable, source_ref=internal_source_ref
                    ),
                    right=ExpressionVariableRef(
                        variable=star_arg_list_variable, source_ref=internal_source_ref
                    ),
                    source_ref=internal_source_ref,
                ),
                kw=ExpressionVariableRef(
                    variable=kw_variable, source_ref=internal_source_ref
                ),
                source_ref=internal_source_ref,
            ),
            source_ref=internal_source_ref,
        ),
    )

    result.setChild("body", body)

    return result


@once_decorator
def getFunctionCallHelperStarDict():
    helper_name = "complex_call_helper_star_dict"
    # Equivalent of:
    #
    # Note: Call in here is not the same, as it can go without checks directly
    # to PyObject_Call.
    #
    # if not isinstance(star_arg_dict, dict):
    #     try:
    #         tmp_keys =  star_arg_dict.keys()
    #     except AttributeError:
    #         raise TypeError, ""%s argument after ** must be a mapping, not %s" % (
    #             get_callable_name_desc(function),
    #             type(star_arg_dict).__name__
    #         )
    #
    #     tmp_iter = iter(keys)
    #     tmp_dict = {}
    #
    #     while 1:
    #         try:
    #             tmp_key = tmp_iter.next()
    #         except StopIteration:
    #             break
    #
    #         tmp_dict[tmp_key] = star_dict_arg[tmp_key]
    #
    #     star_arg_dict = new
    #
    # return called(**star_arg_dict)

    # Only need to check if the star argument value is a sequence and then
    # convert to tuple.
    result = makeInternalHelperFunctionBody(
        name=helper_name,
        parameters=ParameterSpec(
            ps_name=helper_name,
            ps_normal_args=("called", "star_arg_dict"),
            ps_list_star_arg=None,
            ps_dict_star_arg=None,
            ps_default_count=0,
            ps_kw_only_args=(),
            ps_pos_only_args=(),
        ),
    )

    called_variable = result.getVariableForAssignment(variable_name="called")

    star_arg_dict_variable = result.getVariableForAssignment(
        variable_name="star_arg_dict"
    )

    body = makeStatementsSequenceFromStatements(
        _makeStarDictArgumentToDictStatement(
            result=result,
            called_variable=called_variable,
            star_dict_variable=star_arg_dict_variable,
        ),
        StatementReturn(
            expression=makeExpressionCall(
                called=ExpressionVariableRef(
                    variable=called_variable, source_ref=internal_source_ref
                ),
                args=None,
                kw=ExpressionVariableRef(
                    variable=star_arg_dict_variable, source_ref=internal_source_ref
                ),
                source_ref=internal_source_ref,
            ),
            source_ref=internal_source_ref,
        ),
    )

    result.setChild("body", body)

    return result


@once_decorator
def getFunctionCallHelperPosStarDict():
    helper_name = "complex_call_helper_pos_star_dict"

    # Equivalent of:
    #
    # Note: Call in here is not the same, as it can go without checks directly
    # to PyObject_Call.
    #
    # if not isinstance(star_arg_dict, dict):
    #     try:
    #         tmp_keys =  star_arg_dict.keys()
    #     except AttributeError:
    #         raise TypeError, ""%s argument after ** must be a mapping, not %s" % (
    #             get_callable_name_desc(function),
    #             type(star_arg_dict).__name__
    #         )
    #
    #     tmp_iter = iter(keys)
    #     tmp_dict = {}
    #
    #     while 1:
    #         try:
    #             tmp_key = tmp_iter.next()
    #         except StopIteration:
    #             break
    #
    #         tmp_dict[tmp_key] = star_dict_arg[tmp_key]
    #
    #     star_arg_dict = new
    #
    # return called(args, **star_arg_dict)

    # Only need to check if the star argument value is a sequence and then
    # convert to tuple.
    result = makeInternalHelperFunctionBody(
        name=helper_name,
        parameters=ParameterSpec(
            ps_name=helper_name,
            ps_normal_args=("called", "args", "star_arg_dict"),
            ps_list_star_arg=None,
            ps_dict_star_arg=None,
            ps_default_count=0,
            ps_kw_only_args=(),
            ps_pos_only_args=(),
        ),
    )

    called_variable = result.getVariableForAssignment(variable_name="called")

    args_variable = result.getVariableForAssignment(variable_name="args")

    star_arg_dict_variable = result.getVariableForAssignment(
        variable_name="star_arg_dict"
    )

    body = makeStatementsSequenceFromStatements(
        _makeStarDictArgumentToDictStatement(
            result=result,
            called_variable=called_variable,
            star_dict_variable=star_arg_dict_variable,
        ),
        StatementReturn(
            expression=makeExpressionCall(
                called=ExpressionVariableRef(
                    variable=called_variable, source_ref=internal_source_ref
                ),
                args=ExpressionVariableRef(
                    variable=args_variable, source_ref=internal_source_ref
                ),
                kw=ExpressionVariableRef(
                    variable=star_arg_dict_variable, source_ref=internal_source_ref
                ),
                source_ref=internal_source_ref,
            ),
            source_ref=internal_source_ref,
        ),
    )

    result.setChild("body", body)

    return result


@once_decorator
def getFunctionCallHelperKeywordsStarDict():
    helper_name = "complex_call_helper_keywords_star_dict"
    # Equivalent of:
    #
    # Note: Call in here is not the same, as it can go without checks directly
    # to PyObject_Call. One goal is to avoid copying "kw" unless really
    # necessary, and to take the slow route only for non-dictionaries.
    #
    # if not isinstance(star_arg_dict, dict):
    #     try:
    #         tmp_keys =  star_arg_dict.keys()
    #     except AttributeError:
    #         raise TypeError, ""%s argument after ** must be a mapping, not %s" % (
    #             get_callable_name_desc(function),
    #             type(star_arg_dict).__name__
    #         )
    #
    #     if keys:
    #         kw = dict(kw)
    #
    #         tmp_iter = iter(keys)
    #         tmp_dict = {}
    #
    #         while 1:
    #             try:
    #                 tmp_key = tmp_iter.next()
    #             except StopIteration:
    #                  break
    #
    #             if tmp_key in kw:
    #                 raise TypeError, "%s got multiple values for keyword argument '%s'" % (
    #                     get_callable_name_desc(function),
    #                     tmp_key
    #                 )
    #
    #             kw[tmp_key] = star_dict_arg[tmp_key)
    #
    # elif star_arg_dict:
    #    tmp_iter = star_arg_dict.iteritems()
    #
    #    kw = dict(kw)
    #    while 1:
    #        try:
    #            tmp_key, tmp_value = tmp_iter.next()
    #        except StopIteration:
    #            break
    #
    #        if tmp_key in kw:
    #            raise TypeError, "%s got multiple values for keyword argument '%s'" % (
    #                 get_callable_name_desc(function),
    #                 tmp_key
    #            )
    #
    #        kw[tmp_key] = tmp_value
    #
    # return called(**kw)

    # Only need to check if the star argument value is a sequence and then
    # convert to tuple.
    result = makeInternalHelperFunctionBody(
        name=helper_name,
        parameters=ParameterSpec(
            ps_name=helper_name,
            ps_normal_args=("called", "kw", "star_arg_dict"),
            ps_list_star_arg=None,
            ps_dict_star_arg=None,
            ps_default_count=0,
            ps_kw_only_args=(),
            ps_pos_only_args=(),
        ),
    )

    called_variable = result.getVariableForAssignment(variable_name="called")

    kw_variable = result.getVariableForAssignment(variable_name="kw")

    star_arg_dict_variable = result.getVariableForAssignment(
        variable_name="star_arg_dict"
    )

    body = makeStatementsSequenceFromStatements(
        _makeStarDictArgumentMergeToKwStatement(
            result=result,
            called_variable=called_variable,
            kw_variable=kw_variable,
            star_dict_variable=star_arg_dict_variable,
        ),
        StatementReturn(
            expression=makeExpressionCall(
                called=ExpressionVariableRef(
                    variable=called_variable, source_ref=internal_source_ref
                ),
                args=None,
                kw=ExpressionVariableRef(
                    variable=kw_variable, source_ref=internal_source_ref
                ),
                source_ref=internal_source_ref,
            ),
            source_ref=internal_source_ref,
        ),
    )

    result.setChild("body", body)

    return result


@once_decorator
def getFunctionCallHelperPosKeywordsStarDict():
    helper_name = "complex_call_helper_pos_keywords_star_dict"
    # Equivalent of:
    #
    # Note: Call in here is not the same, as it can go without checks directly
    # to PyObject_Call. One goal is to avoid copying "kw" unless really
    # necessary, and to take the slow route only for non-dictionaries.
    #
    # if not isinstance(star_arg_dict, dict):
    #     try:
    #         tmp_keys =  star_arg_dict.keys()
    #     except AttributeError:
    #         raise TypeError, ""%s argument after ** must be a mapping, not %s" % (
    #             get_callable_name_desc(function),
    #             type(star_arg_dict).__name__
    #         )
    #
    #     if keys:
    #         kw = dict(kw)
    #
    #         tmp_iter = iter(keys)
    #         tmp_dict = {}
    #
    #         while 1:
    #             try:
    #                 tmp_key = tmp_iter.next()
    #             except StopIteration:
    #                  break
    #
    #             if tmp_key in kw:
    #                 raise TypeError, "%s got multiple values for keyword argument '%s'" % (
    #                     get_callable_name_desc(function),
    #                     tmp_key
    #                 )
    #
    #             kw[tmp_key] = star_dict_arg[tmp_key]
    #
    # elif star_arg_dict:
    #    tmp_iter = star_arg_dict.iteritems()
    #
    #    kw = dict(kw)
    #    while 1:
    #        try:
    #            tmp_key, tmp_value = tmp_iter.next()
    #        except StopIteration:
    #            break
    #
    #        if tmp_key in kw:
    #            raise TypeError, "%s got multiple values for keyword argument '%s'" % (
    #                 get_callable_name_desc(function),
    #                 tmp_key
    #            )
    #
    #        kw[tmp_key] = tmp_value
    #
    # return called(**kw )

    # Only need to check if the star argument value is a sequence and then
    # convert to tuple.
    result = makeInternalHelperFunctionBody(
        name=helper_name,
        parameters=ParameterSpec(
            ps_name=helper_name,
            ps_normal_args=("called", "args", "kw", "star_arg_dict"),
            ps_list_star_arg=None,
            ps_dict_star_arg=None,
            ps_default_count=0,
            ps_kw_only_args=(),
            ps_pos_only_args=(),
        ),
    )

    called_variable = result.getVariableForAssignment(variable_name="called")

    args_variable = result.getVariableForAssignment(variable_name="args")

    kw_variable = result.getVariableForAssignment(variable_name="kw")

    star_arg_dict_variable = result.getVariableForAssignment(
        variable_name="star_arg_dict"
    )

    body = makeStatementsSequenceFromStatements(
        _makeStarDictArgumentMergeToKwStatement(
            result=result,
            called_variable=called_variable,
            kw_variable=kw_variable,
            star_dict_variable=star_arg_dict_variable,
        ),
        StatementReturn(
            expression=makeExpressionCall(
                called=ExpressionVariableRef(
                    variable=called_variable, source_ref=internal_source_ref
                ),
                args=ExpressionVariableRef(
                    variable=args_variable, source_ref=internal_source_ref
                ),
                kw=ExpressionVariableRef(
                    variable=kw_variable, source_ref=internal_source_ref
                ),
                source_ref=internal_source_ref,
            ),
            source_ref=internal_source_ref,
        ),
    )

    result.setChild("body", body)

    return result


def getDoubleStarArgsConversion(
    result, called_variable, kw_variable, star_arg_list_variable, star_arg_dict_variable
):

    statements = []

    if kw_variable is not None:
        statements.append(
            _makeStarDictArgumentMergeToKwStatement(
                result=result,
                called_variable=called_variable,
                kw_variable=kw_variable,
                star_dict_variable=star_arg_dict_variable,
            )
        )
    else:
        statements.append(
            _makeStarDictArgumentToDictStatement(
                result=result,
                called_variable=called_variable,
                star_dict_variable=star_arg_dict_variable,
            )
        )

    statements.append(
        _makeStarListArgumentToTupleStatement(
            called_variable=called_variable, star_list_variable=star_arg_list_variable
        )
    )

    return statements


@once_decorator
def getFunctionCallHelperStarListStarDict():
    helper_name = "complex_call_helper_star_list_star_dict"

    # Only need to check if the star argument value is a sequence and then
    # convert to tuple.
    result = makeInternalHelperFunctionBody(
        name=helper_name,
        parameters=ParameterSpec(
            ps_name=helper_name,
            ps_normal_args=("called", "star_arg_list", "star_arg_dict"),
            ps_list_star_arg=None,
            ps_dict_star_arg=None,
            ps_default_count=0,
            ps_kw_only_args=(),
            ps_pos_only_args=(),
        ),
    )

    called_variable = result.getVariableForAssignment(variable_name="called")

    star_arg_list_variable = result.getVariableForAssignment(
        variable_name="star_arg_list"
    )

    star_arg_dict_variable = result.getVariableForAssignment(
        variable_name="star_arg_dict"
    )

    statements = getDoubleStarArgsConversion(
        result=result,
        called_variable=called_variable,
        star_arg_list_variable=star_arg_list_variable,
        kw_variable=None,
        star_arg_dict_variable=star_arg_dict_variable,
    )

    statements.append(
        StatementReturn(
            expression=makeExpressionCall(
                called=ExpressionVariableRef(
                    variable=called_variable, source_ref=internal_source_ref
                ),
                args=ExpressionVariableRef(
                    variable=star_arg_list_variable, source_ref=internal_source_ref
                ),
                kw=ExpressionVariableRef(
                    variable=star_arg_dict_variable, source_ref=internal_source_ref
                ),
                source_ref=internal_source_ref,
            ),
            source_ref=internal_source_ref,
        )
    )

    body = makeStatementsSequenceFromStatements(*statements)

    result.setChild("body", body)

    return result


@once_decorator
def getFunctionCallHelperPosStarListStarDict():
    helper_name = "complex_call_helper_pos_star_list_star_dict"

    # Only need to check if the star argument value is a sequence and then
    # convert to tuple.
    result = makeInternalHelperFunctionBody(
        name=helper_name,
        parameters=ParameterSpec(
            ps_name=helper_name,
            ps_normal_args=("called", "args", "star_arg_list", "star_arg_dict"),
            ps_list_star_arg=None,
            ps_dict_star_arg=None,
            ps_default_count=0,
            ps_kw_only_args=(),
            ps_pos_only_args=(),
        ),
    )

    called_variable = result.getVariableForAssignment(variable_name="called")

    args_variable = result.getVariableForAssignment(variable_name="args")

    star_arg_list_variable = result.getVariableForAssignment(
        variable_name="star_arg_list"
    )

    star_arg_dict_variable = result.getVariableForAssignment(
        variable_name="star_arg_dict"
    )

    statements = getDoubleStarArgsConversion(
        result=result,
        called_variable=called_variable,
        star_arg_list_variable=star_arg_list_variable,
        kw_variable=None,
        star_arg_dict_variable=star_arg_dict_variable,
    )

    if python_version >= 0x360:
        statements.reverse()

    statements.append(
        StatementReturn(
            expression=makeExpressionCall(
                called=ExpressionVariableRef(
                    variable=called_variable, source_ref=internal_source_ref
                ),
                args=makeBinaryOperationNode(
                    operator="Add",
                    left=ExpressionVariableRef(
                        variable=args_variable, source_ref=internal_source_ref
                    ),
                    right=ExpressionVariableRef(
                        variable=star_arg_list_variable, source_ref=internal_source_ref
                    ),
                    source_ref=internal_source_ref,
                ),
                kw=ExpressionVariableRef(
                    variable=star_arg_dict_variable, source_ref=internal_source_ref
                ),
                source_ref=internal_source_ref,
            ),
            source_ref=internal_source_ref,
        )
    )

    body = makeStatementsSequenceFromStatements(*statements)

    result.setChild("body", body)

    return result


@once_decorator
def getFunctionCallHelperKeywordsStarListStarDict():
    helper_name = "complex_call_helper_keywords_star_list_star_dict"

    # Only need to check if the star argument value is a sequence and then
    # convert to tuple.
    result = makeInternalHelperFunctionBody(
        name=helper_name,
        parameters=ParameterSpec(
            ps_name=helper_name,
            ps_normal_args=orderArgs("called", "kw", "star_arg_list", "star_arg_dict"),
            ps_list_star_arg=None,
            ps_dict_star_arg=None,
            ps_default_count=0,
            ps_kw_only_args=(),
            ps_pos_only_args=(),
        ),
    )

    called_variable = result.getVariableForAssignment(variable_name="called")

    kw_variable = result.getVariableForAssignment(variable_name="kw")

    star_arg_list_variable = result.getVariableForAssignment(
        variable_name="star_arg_list"
    )

    star_arg_dict_variable = result.getVariableForAssignment(
        variable_name="star_arg_dict"
    )

    statements = getDoubleStarArgsConversion(
        result=result,
        called_variable=called_variable,
        star_arg_list_variable=star_arg_list_variable,
        kw_variable=kw_variable,
        star_arg_dict_variable=star_arg_dict_variable,
    )

    statements.append(
        StatementReturn(
            expression=makeExpressionCall(
                called=ExpressionVariableRef(
                    variable=called_variable, source_ref=internal_source_ref
                ),
                args=ExpressionVariableRef(
                    variable=star_arg_list_variable, source_ref=internal_source_ref
                ),
                kw=ExpressionVariableRef(
                    variable=kw_variable, source_ref=internal_source_ref
                ),
                source_ref=internal_source_ref,
            ),
            source_ref=internal_source_ref,
        )
    )

    body = makeStatementsSequenceFromStatements(*statements)

    result.setChild("body", body)

    return result


@once_decorator
def getFunctionCallHelperPosKeywordsStarListStarDict():
    helper_name = "complex_call_helper_pos_keywords_star_list_star_dict"

    # Only need to check if the star argument value is a sequence and then
    # convert to tuple.
    result = makeInternalHelperFunctionBody(
        name=helper_name,
        parameters=ParameterSpec(
            ps_name=helper_name,
            ps_normal_args=orderArgs(
                "called", "args", "kw", "star_arg_list", "star_arg_dict"
            ),
            ps_list_star_arg=None,
            ps_dict_star_arg=None,
            ps_default_count=0,
            ps_kw_only_args=(),
            ps_pos_only_args=(),
        ),
    )

    called_variable = result.getVariableForAssignment(variable_name="called")

    args_variable = result.getVariableForAssignment(variable_name="args")

    kw_variable = result.getVariableForAssignment(variable_name="kw")

    star_arg_list_variable = result.getVariableForAssignment(
        variable_name="star_arg_list"
    )

    star_arg_dict_variable = result.getVariableForAssignment(
        variable_name="star_arg_dict"
    )

    statements = getDoubleStarArgsConversion(
        result=result,
        called_variable=called_variable,
        star_arg_list_variable=star_arg_list_variable,
        kw_variable=kw_variable,
        star_arg_dict_variable=star_arg_dict_variable,
    )

    if python_version >= 0x360:
        statements.reverse()

    statements.append(
        StatementReturn(
            expression=makeExpressionCall(
                called=ExpressionVariableRef(
                    variable=called_variable, source_ref=internal_source_ref
                ),
                args=makeBinaryOperationNode(
                    operator="Add",
                    left=ExpressionVariableRef(
                        variable=args_variable, source_ref=internal_source_ref
                    ),
                    right=ExpressionVariableRef(
                        variable=star_arg_list_variable, source_ref=internal_source_ref
                    ),
                    source_ref=internal_source_ref,
                ),
                kw=ExpressionVariableRef(
                    variable=kw_variable, source_ref=internal_source_ref
                ),
                source_ref=internal_source_ref,
            ),
            source_ref=internal_source_ref,
        )
    )

    body = makeStatementsSequenceFromStatements(*statements)

    result.setChild("body", body)

    return result


@once_decorator
def getFunctionCallHelperDictionaryUnpacking():
    helper_name = "complex_call_helper_dict_unpacking_checks"

    result = makeInternalHelperFunctionBody(
        name=helper_name,
        parameters=ParameterSpec(
            ps_name=helper_name,
            ps_normal_args=("called",),
            ps_list_star_arg="args",
            ps_dict_star_arg=None,
            ps_default_count=0,
            ps_kw_only_args=(),
            ps_pos_only_args=(),
        ),
    )

    args_variable = result.getVariableForAssignment(variable_name="args")
    called_variable = result.getVariableForAssignment(variable_name="called")

    temp_scope = None

    tmp_result_variable = result.allocateTempVariable(temp_scope, "dict")
    tmp_iter_variable = result.allocateTempVariable(temp_scope, "dicts_iter")
    tmp_item_variable = result.allocateTempVariable(temp_scope, "args_item")
    tmp_iter2_variable = result.allocateTempVariable(temp_scope, "dict_iter")
    tmp_key_variable = result.allocateTempVariable(temp_scope, "dict_key")

    update_body = (
        makeStatementConditional(
            condition=ExpressionComparisonIsNot(
                left=ExpressionBuiltinType1(
                    value=ExpressionTempVariableRef(
                        variable=tmp_key_variable, source_ref=internal_source_ref
                    ),
                    source_ref=internal_source_ref,
                ),
                right=makeExpressionBuiltinTypeRef(
                    builtin_name="str", source_ref=internal_source_ref
                ),
                source_ref=internal_source_ref,
            ),
            yes_branch=_makeRaiseNoStringItem(called_variable=called_variable),
            no_branch=None,
            source_ref=internal_source_ref,
        ),
        makeStatementConditional(
            condition=ExpressionComparisonIn(
                left=ExpressionTempVariableRef(
                    variable=tmp_key_variable, source_ref=internal_source_ref
                ),
                right=ExpressionTempVariableRef(
                    variable=tmp_result_variable, source_ref=internal_source_ref
                ),
                source_ref=internal_source_ref,
            ),
            yes_branch=_makeRaiseDuplicationItem(
                called_variable=called_variable, tmp_key_variable=tmp_key_variable
            ),
            no_branch=None,
            source_ref=internal_source_ref,
        ),
        StatementDictOperationSetKeyValue(
            dict_arg=ExpressionTempVariableRef(
                variable=tmp_result_variable, source_ref=internal_source_ref
            ),
            key=ExpressionTempVariableRef(
                variable=tmp_key_variable, source_ref=internal_source_ref
            ),
            value=ExpressionSubscriptLookup(
                expression=ExpressionTempVariableRef(
                    variable=tmp_item_variable, source_ref=internal_source_ref
                ),
                subscript=ExpressionTempVariableRef(
                    variable=tmp_key_variable, source_ref=internal_source_ref
                ),
                source_ref=internal_source_ref,
            ),
            source_ref=internal_source_ref,
        ),
    )

    loop_body = (
        makeTryExceptSingleHandlerNode(
            tried=makeStatementsSequenceFromStatements(
                makeStatementAssignmentVariable(
                    variable=tmp_iter2_variable,
                    source=ExpressionBuiltinIter1(
                        value=makeCallNode(
                            _makeNameAttributeLookup(
                                ExpressionTempVariableRef(
                                    variable=tmp_item_variable,
                                    source_ref=internal_source_ref,
                                ),
                                attribute_name="keys",
                            ),
                            internal_source_ref,
                        ),
                        source_ref=internal_source_ref,
                    ),
                    source_ref=internal_source_ref,
                ),
                _makeIteratingLoopStatement(
                    tmp_iter_variable=tmp_iter2_variable,
                    tmp_item_variable=tmp_key_variable,
                    statements=update_body,
                ),
            ),
            exception_name="AttributeError",
            handler_body=_makeRaiseExceptionMustBeMapping(
                called_variable=called_variable, star_dict_variable=tmp_item_variable
            ),
            source_ref=internal_source_ref,
        ),
    )

    final = (
        StatementReleaseVariable(
            variable=tmp_result_variable, source_ref=internal_source_ref
        ),
        StatementReleaseVariable(
            variable=tmp_iter_variable, source_ref=internal_source_ref
        ),
        StatementReleaseVariable(
            variable=tmp_item_variable, source_ref=internal_source_ref
        ),
        StatementReleaseVariable(
            variable=tmp_iter2_variable, source_ref=internal_source_ref
        ),
        StatementReleaseVariable(
            variable=tmp_key_variable, source_ref=internal_source_ref
        ),
    )

    tried = makeStatementsSequenceFromStatements(
        makeStatementAssignmentVariable(
            variable=tmp_iter_variable,
            source=ExpressionBuiltinIter1(
                value=ExpressionVariableRef(
                    variable=args_variable, source_ref=internal_source_ref
                ),
                source_ref=internal_source_ref,
            ),
            source_ref=internal_source_ref,
        ),
        makeStatementAssignmentVariable(
            variable=tmp_result_variable,
            source=makeConstantRefNode(constant={}, source_ref=internal_source_ref),
            source_ref=internal_source_ref,
        ),
        _makeIteratingLoopStatement(
            tmp_iter_variable=tmp_iter_variable,
            tmp_item_variable=tmp_item_variable,
            statements=loop_body,
        ),
        StatementReturn(
            expression=ExpressionTempVariableRef(
                variable=tmp_result_variable, source_ref=internal_source_ref
            ),
            source_ref=internal_source_ref,
        ),
    )

    body = makeStatementsSequenceFromStatement(
        makeTryFinallyStatement(
            provider=result,
            tried=tried,
            final=final,
            source_ref=internal_source_ref,
        )
    )

    result.setChild("body", body)

    return result
